import json
import logging
import re
from builtins import str  # PY2 compatibility

import altair as alt
import pandas as pd

import traceback

from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationValidationResult,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.render.renderer.content_block import (
    ExceptionListContentBlockRenderer,
)
from great_expectations.render.renderer.renderer import Renderer
from great_expectations.render.types import (
    RenderedBulletListContent,
    RenderedGraphContent,
    RenderedHeaderContent,
    RenderedSectionContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    TextContent,
    ValueListContent,
)
from great_expectations.util import (
    load_class,
    verify_dynamic_loading_support,
)
from great_expectations.exceptions import (
    ClassInstantiationError,
    GreatExpectationsError,
)

logger = logging.getLogger(__name__)


def convert_to_string_and_escape(var):
    return re.sub(r"\$", r"$$", str(var))


class ColumnSectionRenderer(Renderer):
    @classmethod
    def _get_column_name(cls, ge_object):
        # This is broken out for ease of locating future validation here
        if isinstance(ge_object, list):
            candidate_object = ge_object[0]
        else:
            candidate_object = ge_object
        try:
            if isinstance(candidate_object, ExpectationConfiguration):
                return candidate_object.kwargs["column"]
            elif isinstance(candidate_object, ExpectationValidationResult):
                return candidate_object.expectation_config.kwargs["column"]
            else:
                raise ValueError(
                    "Provide a column section renderer an expectation, list of expectations, evr, or list of evrs.")
        except KeyError:
            return "Table-Level Expectations"


class ProfilingResultsColumnSectionRenderer(ColumnSectionRenderer):

    def __init__(self, overview_table_renderer=None, expectation_string_renderer=None, runtime_environment=None):
        if overview_table_renderer is None:
            overview_table_renderer = {
                "class_name": "ProfilingOverviewTableContentBlockRenderer"
            }
        if expectation_string_renderer is None:
            expectation_string_renderer = {
                "class_name": "ExpectationStringRenderer"
            }
        module_name = 'great_expectations.render.renderer.content_block'
        self._overview_table_renderer = instantiate_class_from_config(
            config=overview_table_renderer,
            runtime_environment=runtime_environment,
            config_defaults={
                "module_name": module_name
            }
        )
        if not self._overview_table_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=overview_table_renderer['class_name']
            )
        self._expectation_string_renderer = instantiate_class_from_config(
            config=expectation_string_renderer,
            runtime_environment=runtime_environment,
            config_defaults={
                "module_name": module_name
            }
        )
        if not self._expectation_string_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=expectation_string_renderer['class_name']
            )

        self.content_block_function_names = [
            "_render_header",
            "_render_overview_table",
            "_render_quantile_table",
            "_render_stats_table",
            "_render_values_set",
            "_render_histogram",
            "_render_bar_chart_table",
            "_render_failed",
        ]

    #Note: Seems awkward to pass section_name and column_type into this renderer.
    #Can't we figure that out internally?
    def render(self, evrs, section_name=None, column_type=None):
        if section_name is None:
            column = self._get_column_name(evrs)
        else:
            column = section_name

        content_blocks = []

        for content_block_function_name in self.content_block_function_names:
            try:
                if content_block_function_name == "_render_header":
                    content_blocks.append(getattr(self, content_block_function_name)(evrs, column_type))
                else:
                    content_blocks.append(getattr(self, content_block_function_name)(evrs))
            except Exception as e:
                exception_message = f'''\
An unexpected Exception occurred during data docs rendering.  Because of this error, certain parts of data docs will \
not be rendered properly and/or may not appear altogether.  Please use the trace, included in this message, to \
diagnose and repair the underlying issue.  Detailed information follows:  
                '''
                exception_traceback = traceback.format_exc()
                exception_message += f'{type(e).__name__}: "{str(e)}".  Traceback: "{exception_traceback}".'
                logger.error(exception_message, e, exc_info=True)

        # NOTE : Some render* functions return None so we filter them out
        populated_content_blocks = list(filter(None, content_blocks))

        return RenderedSectionContent(**{
            "section_name": column,
            "content_blocks": populated_content_blocks,
        })

    @classmethod
    def _render_header(cls, evrs, column_type=None):
        # NOTE: This logic is brittle
        try:
            column_name = evrs[0].expectation_config.kwargs["column"]
        except KeyError:
            column_name = "Table-level expectations"

        return RenderedHeaderContent(**{
            "content_block_type": "header",
            "header": RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": convert_to_string_and_escape(column_name),
                            "tooltip": {
                                "content": "expect_column_to_exist",
                                "placement": "top"
                            },
                            "tag": "h5",
                            "styling": {
                                "classes": ["m-0", "p-0"]
                            }
                        }
                    }),
            "subheader": RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "Type: {column_type}".format(column_type=column_type),
                            "tooltip": {
                              "content":
                              "expect_column_values_to_be_of_type <br>expect_column_values_to_be_in_type_list",
                            },
                            "tag": "h6",
                            "styling": {
                                "classes": ["mt-1", "mb-0"]
                            }
                        }
                    }),
            # {
            #     "template": column_type,
            # },
            "styling": {
                "classes": ["col-12", "p-0"],
                "header": {
                    "classes": ["alert", "alert-secondary"]
                }
            }
        })

    @classmethod
    def _render_expectation_types(cls, evrs, content_blocks):
        # NOTE: The evr-fetching function is an kinda similar to the code other_section_
        # renderer.ProfilingResultsOverviewSectionRenderer._render_expectation_types

        # type_counts = defaultdict(int)

        # for evr in evrs:
        #     type_counts[evr.expectation_config.expectation_type] += 1

        # bullet_list = sorted(type_counts.items(), key=lambda kv: -1*kv[1])

        bullet_list = [{
            "content_block_type": "string_template",
            "string_template": {
                "template": "$expectation_type $is_passing",
                "params": {
                    "expectation_type": evr.expectation_config.expectation_type,
                    "is_passing": str(evr.success),
                },
                "styling": {
                    "classes": ["list-group-item", "d-flex", "justify-content-between", "align-items-center"],
                    "params": {
                        "is_passing": {
                            "classes": ["badge", "badge-secondary", "badge-pill"],
                        }
                    },
                }
            }
        } for evr in evrs]

        content_blocks.append(RenderedBulletListContent(**{
            "content_block_type": "bullet_list",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Expectation types <span class="mr-3 triangle"></span>',
                    "tag": "h6"
                }
            }),
            "bullet_list": bullet_list,
            "styling": {
                "classes": ["col-12", "mt-1"],
                "header": {
                    "classes": ["collapsed"],
                    "attributes": {
                        "data-toggle": "collapse",
                        "href": "#{{content_block_id}}-body",
                        "role": "button",
                        "aria-expanded": "true",
                        "aria-controls": "collapseExample",
                    },
                    "styles": {
                        "cursor": "pointer",
                    }
                },
                "body": {
                    "classes": ["list-group", "collapse"],
                },
            },
        }))

    def _render_overview_table(self, evrs):
        unique_n = self._find_evr_by_type(
            evrs,
            "expect_column_unique_value_count_to_be_between"
        )
        unique_proportion = self._find_evr_by_type(
            evrs,
            "expect_column_proportion_of_unique_values_to_be_between"
        )
        null_evr = self._find_evr_by_type(
            evrs,
            "expect_column_values_to_not_be_null"
        )
        evrs = [evr for evr in [unique_n, unique_proportion, null_evr] if (evr is not None)]

        if len(evrs) > 0:
            new_content_block = self._overview_table_renderer.render(evrs)
            new_content_block.header = RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Properties',
                    "tag": "h6"
                }
            })
            new_content_block.styling = {
                "classes": ["col-3", "mt-1", "pl-1", "pr-1"],
                "body": {
                    "classes": ["table", "table-sm", "table-unbordered"],
                    "styles": {
                        "width": "100%"
                    },
                }

            }
            return new_content_block

    @classmethod
    def _render_quantile_table(cls, evrs):
        table_rows = []

        quantile_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_quantile_values_to_be_between"
        )

        if not quantile_evr or quantile_evr.exception_info["raised_exception"]:
            return

        quantiles = quantile_evr.result["observed_value"]["quantiles"]
        quantile_ranges = quantile_evr.result["observed_value"]["values"]

        quantile_strings = {
            .25: "Q1",
            .75: "Q3",
            .50: "Median"
        }

        for idx, quantile in enumerate(quantiles):
            quantile_string = quantile_strings.get(quantile)
            table_rows.append([
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": quantile_string if quantile_string else "{:3.2f}".format(quantile),
                        "tooltip": {
                            "content": "expect_column_quantile_values_to_be_between \n expect_column_median_to_be_between" if quantile == 0.50 else "expect_column_quantile_values_to_be_between"
                        }
                    }
                },
                quantile_ranges[idx],
            ])

        return RenderedTableContent(**{
            "content_block_type": "table",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Quantiles',
                    "tag": "h6"
                }
            }),
            "table": table_rows,
            "styling": {
                "classes": ["col-3", "mt-1", "pl-1", "pr-1"],
                "body": {
                    "classes": ["table", "table-sm", "table-unbordered"],
                }
            },
        })

    @classmethod
    def _render_stats_table(cls, evrs):
        table_rows = []

        mean_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_mean_to_be_between"
        )

        if not mean_evr or mean_evr.exception_info["raised_exception"]:
            return

        mean_value = "{:.2f}".format(
            mean_evr.result['observed_value']) if mean_evr else None
        if mean_value:
            table_rows.append([
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Mean",
                        "tooltip": {
                            "content": "expect_column_mean_to_be_between"
                        }
                    }
                },
                mean_value
            ])

        min_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_min_to_be_between"
        )
        min_value = "{:.2f}".format(
            min_evr.result['observed_value']) if min_evr else None
        if min_value:
            table_rows.append([
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Minimum",
                        "tooltip": {
                            "content": "expect_column_min_to_be_between"
                        }
                    }
                },
                min_value,
            ])

        max_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_max_to_be_between"
        )
        max_value = "{:.2f}".format(
            max_evr.result['observed_value']) if max_evr else None
        if max_value:
            table_rows.append([
                {
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "Maximum",
                        "tooltip": {
                            "content": "expect_column_max_to_be_between"
                        }
                    }
                },
                max_value
            ])

        if len(table_rows) > 0:
            return RenderedTableContent(**{
                "content_block_type": "table",
                "header": RenderedStringTemplateContent(**{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": 'Statistics',
                        "tag": "h6"
                    }
                }),
                "table": table_rows,
                "styling": {
                    "classes": ["col-3", "mt-1", "pl-1", "pr-1"],
                    "body": {
                        "classes": ["table", "table-sm", "table-unbordered"],
                    }
                },
            })
        else:
            return

    @classmethod
    def _render_values_set(cls, evrs):
        set_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_values_to_be_in_set"
        )

        if not set_evr or set_evr.exception_info["raised_exception"]:
            return

        if set_evr and "partial_unexpected_counts" in set_evr.result:
            partial_unexpected_counts = set_evr.result["partial_unexpected_counts"]
            values = [str(v["value"]) for v in partial_unexpected_counts]
        elif set_evr and "partial_unexpected_list" in set_evr.result:
            values = [str(item) for item in set_evr.result["partial_unexpected_list"]]
        else:
            return

        classes = ["col-3", "mt-1", "pl-1", "pr-1"]

        if any(len(value) > 80 for value in values):
            content_block_type = "bullet_list"
            content_block_class = RenderedBulletListContent
        else:
            content_block_type = "value_list"
            content_block_class = ValueListContent

        new_block = content_block_class(**{
            "content_block_type": content_block_type,
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "Example Values",
                    "tooltip": {
                        "content": "expect_column_values_to_be_in_set"
                    },
                    "tag": "h6"
                }
            }),
            content_block_type: [{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "$value",
                    "params": {
                        "value": value
                    },
                    "styling": {
                        "default": {
                            "classes": ["badge", "badge-info"] if content_block_type == "value_list" else [],
                            "styles": {
                                "word-break": "break-all"
                            }
                        },
                    }
                }
            } for value in values],
            "styling": {
                "classes": classes,
            }
        })

        return new_block

    def _render_histogram(self, evrs):
        # NOTE: This code is very brittle
        kl_divergence_evr = self._find_evr_by_type(
            evrs,
            "expect_column_kl_divergence_to_be_less_than"
        )
        # print(json.dumps(kl_divergence_evr, indent=2))
        if kl_divergence_evr is None or kl_divergence_evr.result is None or "details" not in kl_divergence_evr.result:
            return

        observed_partition_object = kl_divergence_evr.result["details"]["observed_partition"]
        weights = observed_partition_object["weights"]
        if len(weights) > 60:
            return None

        header = RenderedStringTemplateContent(**{
            "content_block_type": "string_template",
            "string_template": {
                "template": "Histogram",
                "tooltip": {
                    "content": "expect_column_kl_divergence_to_be_less_than"
                },
                "tag": "h6"
            }
        })

        return self._expectation_string_renderer._get_kl_divergence_chart(observed_partition_object, header)

    @classmethod
    def _render_bar_chart_table(cls, evrs):
        distinct_values_set_evr = cls._find_evr_by_type(
            evrs,
            "expect_column_distinct_values_to_be_in_set"
        )
        if not distinct_values_set_evr or distinct_values_set_evr.exception_info["raised_exception"]:
            return

        value_count_dicts = distinct_values_set_evr.result['details']['value_counts']
        if isinstance(value_count_dicts, pd.Series):
            values = value_count_dicts.index.tolist()
            counts = value_count_dicts.tolist()
        else:
            values = [value_count_dict['value'] for value_count_dict in value_count_dicts]
            counts = [value_count_dict['count'] for value_count_dict in value_count_dicts]

        df = pd.DataFrame({
            "value": values,
            "count": counts,
        })

        if len(values) > 60:
            return None
        else:
            chart_pixel_width = (len(values) / 60.0) * 500
            if chart_pixel_width < 250:
                chart_pixel_width = 250
            chart_container_col_width = round((len(values) / 60.0) * 6)
            if chart_container_col_width < 4:
                chart_container_col_width = 4
            elif chart_container_col_width >= 5:
                chart_container_col_width = 6
            elif chart_container_col_width >= 4:
                chart_container_col_width = 5

        mark_bar_args = {}
        if len(values) == 1:
            mark_bar_args["size"] = 20

        bars = alt.Chart(df).mark_bar(**mark_bar_args).encode(
            y='count:Q',
            x="value:O",
            tooltip=["value", "count"]
        ).properties(height=400, width=chart_pixel_width, autosize="fit")

        chart = bars.to_json()

        new_block = RenderedGraphContent(**{
            "content_block_type": "graph",
            "header": RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "Value Counts",
                            "tooltip": {
                                "content": "expect_column_distinct_values_to_be_in_set"
                            },
                            "tag": "h6"
                        }
                    }),
            "graph": chart,
            "styling": {
                "classes": ["col-" + str(chart_container_col_width), "mt-1"],
            }
        })

        return new_block

    @classmethod
    def _render_failed(cls, evrs):
        return ExceptionListContentBlockRenderer.render(evrs, include_column_name=False)

    @classmethod
    def _render_unrecognized(cls, evrs, content_blocks):
        unrendered_blocks = []
        new_block = None
        for evr in evrs:
            if evr.expectation_config.expectation_type not in [
                "expect_column_to_exist",
                "expect_column_values_to_be_of_type",
                "expect_column_values_to_be_in_set",
                "expect_column_unique_value_count_to_be_between",
                "expect_column_proportion_of_unique_values_to_be_between",
                "expect_column_values_to_not_be_null",
                "expect_column_max_to_be_between",
                "expect_column_mean_to_be_between",
                "expect_column_min_to_be_between"
            ]:
                new_block = TextContent(**{
                    "content_block_type": "text",
                    "text": []
                })
                new_block["content"].append("""
    <div class="alert alert-primary" role="alert">
        Warning! Unrendered EVR:<br/>
    <pre>"""+json.dumps(evr, indent=2)+"""</pre>
    </div>
                """)

        if new_block is not None:
            unrendered_blocks.append(new_block)

        # print(unrendered_blocks)
        content_blocks += unrendered_blocks


class ValidationResultsColumnSectionRenderer(ColumnSectionRenderer):

    def __init__(self, table_renderer=None):
        if table_renderer is None:
            table_renderer = {
                "class_name": "ValidationResultsTableContentBlockRenderer"
            }
        module_name = table_renderer.get("module_name", "great_expectations.render.renderer.content_block")
        verify_dynamic_loading_support(module_name=module_name, package_name=None)
        class_name = table_renderer.get("class_name")
        self._table_renderer = load_class(
            class_name=class_name,
            module_name=module_name
        )

    @classmethod
    def _render_header(cls, validation_results):
        column = cls._get_column_name(validation_results)

        new_block = RenderedHeaderContent(**{
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": convert_to_string_and_escape(column),
                    "tag": "h5",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "styling": {
                "classes": ["col-12", "p-0"],
                "header": {
                    "classes": ["alert", "alert-secondary"]
                }
            }
        })

        return validation_results, new_block

    def _render_table(self, validation_results):
        new_block = self._table_renderer.render(
            validation_results,
            include_column_name=False
        )

        return [], new_block

    def render(self, validation_results):
        column = self._get_column_name(validation_results)
        content_blocks = []
        remaining_evrs, content_block = self._render_header(validation_results)
        content_blocks.append(content_block)
        remaining_evrs, content_block = self._render_table(remaining_evrs)
        content_blocks.append(content_block)

        return RenderedSectionContent(**{
            "section_name": column,
            "content_blocks": content_blocks
        })


class ExpectationSuiteColumnSectionRenderer(ColumnSectionRenderer):

    def __init__(self, bullet_list_renderer=None):
        if bullet_list_renderer is None:
            bullet_list_renderer = {
                "class_name": "ExpectationSuiteBulletListContentBlockRenderer"
            }
        module_name = bullet_list_renderer.get("module_name", "great_expectations.render.renderer.content_block")
        verify_dynamic_loading_support(module_name=module_name, package_name=None)
        class_name = bullet_list_renderer.get("class_name")
        self._bullet_list_renderer = load_class(
            class_name=class_name,
            module_name=module_name
        )

    @classmethod
    def _render_header(cls, expectations):
        column = cls._get_column_name(expectations)

        new_block = RenderedHeaderContent(**{
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": convert_to_string_and_escape(column),
                    "tag": "h5",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "styling": {
                "classes": ["col-12"],
                "header": {
                    "classes": ["alert", "alert-secondary"]
                }
            }
        })

        return expectations, new_block

    def _render_bullet_list(self, expectations):

        new_block = self._bullet_list_renderer.render(
            expectations,
            include_column_name=False,
        )

        return [], new_block

    def render(self, expectations):
        column = self._get_column_name(expectations)

        content_blocks = []
        remaining_expectations, header_block = self._render_header(expectations)
        content_blocks.append(header_block)
        # remaining_expectations, content_blocks = cls._render_column_type(
        # remaining_expectations, content_blocks)
        remaining_expectations, bullet_block = self._render_bullet_list(remaining_expectations)
        content_blocks.append(bullet_block)

        # NOTE : Some render* functions return None so we filter them out
        populated_content_blocks = list(filter(None, content_blocks))
        return RenderedSectionContent(
            section_name=column,
            content_blocks=populated_content_blocks
        )



import os

from click.testing import CliRunner

from great_expectations.cli import cli
from tests.cli.utils import assert_no_logging_messages_or_tracebacks


def test_project_check_on_missing_ge_dir_guides_user_to_fix(caplog, tmp_path_factory):
    project_dir = str(tmp_path_factory.mktemp("empty_dir"))
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli, ["project", "check-config", "-d", project_dir], catch_exceptions=False
    )
    stdout = result.output
    assert "Checking your config files for validity" in stdout
    assert "Unfortunately, your config appears to be invalid" in stdout
    assert "Error: No great_expectations directory was found here!" in stdout
    assert result.exit_code == 1
    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_project_check_on_valid_project_says_so(caplog, titanic_data_context):
    project_dir = titanic_data_context.root_directory
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli, ["project", "check-config", "-d", project_dir], catch_exceptions=False
    )
    assert "Checking your config files for validity" in result.output
    assert "Your config file appears valid" in result.output
    assert result.exit_code == 0
    assert_no_logging_messages_or_tracebacks(caplog, result)


def test_project_check_on_project_with_missing_config_file_guides_user(
    caplog, titanic_data_context
):
    project_dir = titanic_data_context.root_directory
    # Remove the config file.
    os.remove(os.path.join(project_dir, "great_expectations.yml"))

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        cli, ["project", "check-config", "-d", project_dir], catch_exceptions=False
    )
    assert result.exit_code == 1
    assert "Checking your config files for validity" in result.output
    assert "Unfortunately, your config appears to be invalid" in result.output
    assert_no_logging_messages_or_tracebacks(caplog, result)



import logging
import os

from six import string_types

from collections import OrderedDict

from ...core.id_dict import BatchKwargs
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.render.util import num_to_str
from .renderer import Renderer
from ..types import (
    RenderedDocumentContent,
    RenderedSectionContent,
    RenderedHeaderContent,
    RenderedTableContent,
    TextContent,
    RenderedStringTemplateContent,
    RenderedMarkdownContent,
    CollapseContent
)
from great_expectations.exceptions import ClassInstantiationError

logger = logging.getLogger(__name__)


class ValidationResultsPageRenderer(Renderer):

    def __init__(self, column_section_renderer=None):
        if column_section_renderer is None:
            column_section_renderer = {
                "class_name": "ValidationResultsColumnSectionRenderer"
            }
        module_name = 'great_expectations.render.renderer.column_section_renderer'
        self._column_section_renderer = instantiate_class_from_config(
            config=column_section_renderer,
            runtime_environment={},
            config_defaults={
                "module_name": column_section_renderer.get("module_name", module_name)
            }
        )
        if not self._column_section_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=column_section_renderer['class_name']
            )

    def render(self, validation_results):
        run_id = validation_results.meta['run_id']
        batch_id = BatchKwargs(validation_results.meta['batch_kwargs']).to_id()
        expectation_suite_name = validation_results.meta['expectation_suite_name']
        batch_kwargs = validation_results.meta.get("batch_kwargs")

        # add datasource key to batch_kwargs if missing
        if 'datasource' not in validation_results.meta.get("batch_kwargs", {}):
            # check if expectation_suite_name follows datasource.batch_kwargs_generator.data_asset_name.suite_name pattern
            if len(expectation_suite_name.split('.')) == 4:
                batch_kwargs['datasource'] = expectation_suite_name.split('.')[0]

        # Group EVRs by column
        columns = {}
        for evr in validation_results.results:
            if "column" in evr.expectation_config.kwargs:
                column = evr.expectation_config.kwargs["column"]
            else:
                column = "Table-Level Expectations"

            if column not in columns:
                columns[column] = []
            columns[column].append(evr)

        ordered_columns = Renderer._get_column_list_from_evrs(validation_results)

        overview_content_blocks = [
            self._render_validation_header(validation_results),
            self._render_validation_statistics(validation_results=validation_results),
        ]

        collapse_content_blocks = [self._render_validation_info(validation_results=validation_results)]

        if validation_results["meta"].get("batch_markers"):
            collapse_content_blocks.append(
                self._render_nested_table_from_dict(
                    input_dict=validation_results["meta"].get("batch_markers"),
                    header="Batch Markers"
                )
            )

        if validation_results["meta"].get("batch_kwargs"):
            collapse_content_blocks.append(
                self._render_nested_table_from_dict(
                    input_dict=validation_results["meta"].get("batch_kwargs"),
                    header="Batch Kwargs"
                )
            )

        if validation_results["meta"].get("batch_parameters"):
            collapse_content_blocks.append(
                self._render_nested_table_from_dict(
                    input_dict=validation_results["meta"].get("batch_parameters"),
                    header="Batch Parameters"
                )
            )

        collapse_content_block = CollapseContent(**{
            "collapse_toggle_link": "Show more info...",
            "collapse": collapse_content_blocks,
            "styling": {
                "body": {
                    "classes": ["card", "card-body"]
                },
                "classes": ["col-12", "p-1"]
            }
        })

        overview_content_blocks.append(collapse_content_block)

        sections = [
            RenderedSectionContent(**{
                "section_name": "Overview",
                "content_blocks": overview_content_blocks
            })
        ]

        if "Table-Level Expectations" in columns:
            sections += [
                self._column_section_renderer.render(
                    validation_results=columns["Table-Level Expectations"]
                )
            ]

        sections += [
            self._column_section_renderer.render(
                validation_results=columns[column],
            ) for column in ordered_columns
        ]

        return RenderedDocumentContent(**{
            "renderer_type": "ValidationResultsPageRenderer",
            "page_title": expectation_suite_name + " / " + run_id + " / " + batch_id,
            "batch_kwargs": batch_kwargs,
            "expectation_suite_name": expectation_suite_name,
            "sections": sections,
            "utm_medium": "validation-results-page",
        })

    @classmethod
    def _render_validation_header(cls, validation_results):
        success = validation_results.success
        expectation_suite_name = validation_results.meta['expectation_suite_name']
        expectation_suite_path_components = ['..' for _ in range(len(expectation_suite_name.split('.')) + 2)] \
            + ["expectations"] + expectation_suite_name.split(".")
        expectation_suite_path = os.path.join(*expectation_suite_path_components) + ".html"
        if success:
            success = '<i class="fas fa-check-circle text-success" aria-hidden="true"></i> Succeeded'
        else:
            success = '<i class="fas fa-times text-danger" aria-hidden="true"></i> Failed'
        return RenderedHeaderContent(**{
            "content_block_type": "header",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Overview',
                    "tag": "h5",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "subheader": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "${suite_title} ${expectation_suite_name}\n${status_title} ${success}",
                    "params": {
                        "suite_title": "Expectation Suite:",
                        "status_title": "Status:",
                        "expectation_suite_name": expectation_suite_name,
                        "success": success
                    },
                    "styling": {
                        "params": {
                            "suite_title": {
                                "classes": ["h6"]
                            },
                            "status_title": {
                                "classes": ["h6"]
                            },
                            "expectation_suite_name": {
                                "tag": "a",
                                "attributes": {
                                    "href": expectation_suite_path
                                }
                            }
                        },
                        "classes": ["mb-0", "mt-1"]
                    }
                }
            }),
            "styling": {
                "classes": ["col-12", "p-0"],
                "header": {
                    "classes": ["alert", "alert-secondary"]
                }
            }
        })

    @classmethod
    def _render_validation_info(cls, validation_results):
        run_id = validation_results.meta['run_id']
        ge_version = validation_results.meta["great_expectations.__version__"]

        return RenderedTableContent(**{
            "content_block_type": "table",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Info',
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "table": [
                ["Great Expectations Version", ge_version],
                ["Run ID", run_id]
            ],
            "styling": {
                "classes": ["col-12", "table-responsive", "mt-1"],
                "body": {
                    "classes": ["table", "table-sm"]
                }
            },
        })

    @classmethod
    def _render_nested_table_from_dict(cls, input_dict, header=None, sub_table=False):
        table_rows = []

        for kwarg, value in input_dict.items():
            if not isinstance(value, (dict, OrderedDict)):
                table_row = [
                    RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "$value",
                            "params": {
                                "value": str(kwarg)
                            },
                            "styling": {
                                "default": {
                                    "styles": {
                                        "word-break": "break-all"
                                    }
                                },
                            }
                        },
                        "styling": {
                            "parent": {
                                "classes": ["pr-3"],
                            }
                        }
                    }),
                    RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "$value",
                            "params": {
                                "value": str(value)
                            },
                            "styling": {
                                "default": {
                                    "styles": {
                                        "word-break": "break-all"
                                    }
                                },
                            }
                        },
                        "styling": {
                            "parent": {
                                "classes": [],
                            }
                        }
                    })
                ]
            else:
                table_row = [
                    RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": "$value",
                            "params": {
                                "value": str(kwarg)
                            },
                            "styling": {
                                "default": {
                                    "styles": {
                                        "word-break": "break-all"
                                    }
                                },
                            }
                        },
                        "styling": {
                            "parent": {
                                "classes": ["pr-3"],
                            }
                        }
                    }),
                    cls._render_nested_table_from_dict(value, sub_table=True)
                ]
            table_rows.append(table_row)

        table_rows.sort(key=lambda row: row[0].string_template["params"]["value"])

        if sub_table:
            return RenderedTableContent(**{
                "content_block_type": "table",
                "table": table_rows,
                "styling": {
                    "classes": ["col-6", "table-responsive"],
                    "body": {
                        "classes": ["table", "table-sm", "m-0"]
                    },
                    "parent": {
                        "classes": ["pt-0", "pl-0", "border-top-0"]
                    }
                },
            })
        else:
            return RenderedTableContent(**{
                "content_block_type": "table",
                "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": header,
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
                "table": table_rows,
                "styling": {
                    "classes": ["col-6", "table-responsive", "mt-1"],
                    "body": {
                        "classes": ["table", "table-sm"]
                    }
                },
            })

    @classmethod
    def _render_validation_statistics(cls, validation_results):
        statistics = validation_results["statistics"]
        statistics_dict = OrderedDict([
            ("evaluated_expectations", "Evaluated Expectations"),
            ("successful_expectations", "Successful Expectations"),
            ("unsuccessful_expectations", "Unsuccessful Expectations"),
            ("success_percent", "Success Percent")
        ])
        table_rows = []
        for key, value in statistics_dict.items():
            if statistics.get(key) is not None:
                if key == "success_percent":
                    # table_rows.append([value, "{0:.2f}%".format(statistics[key])])
                    table_rows.append([value, num_to_str(statistics[key], precision=4) + "%"])
                else:
                    table_rows.append([value, statistics[key]])

        return RenderedTableContent(**{
            "content_block_type": "table",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Statistics',
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "table": table_rows,
            "styling": {
                "classes": ["col-6", "table-responsive", "mt-1", "p-1"],
                "body": {
                    "classes": ["table", "table-sm"]
                }
            },
        })


class ExpectationSuitePageRenderer(Renderer):

    def __init__(self, column_section_renderer=None):
        if column_section_renderer is None:
            column_section_renderer = {
                "class_name": "ExpectationSuiteColumnSectionRenderer"
            }
        module_name = 'great_expectations.render.renderer.column_section_renderer'
        self._column_section_renderer = instantiate_class_from_config(
            config=column_section_renderer,
            runtime_environment={},
            config_defaults={
                "module_name": column_section_renderer.get("module_name", module_name)
            }
        )
        if not self._column_section_renderer:
            raise ClassInstantiationError(
                module_name=column_section_renderer,
                package_name=None,
                class_name=column_section_renderer['class_name']
            )

    def render(self, expectations):
        columns, ordered_columns = self._group_and_order_expectations_by_column(expectations)
        expectation_suite_name = expectations.expectation_suite_name

        overview_content_blocks = [
            self._render_expectation_suite_header(),
            self._render_expectation_suite_info(expectations)
        ]

        table_level_expectations_content_block = self._render_table_level_expectations(columns)
        if table_level_expectations_content_block is not None:
            overview_content_blocks.append(table_level_expectations_content_block)

        asset_notes_content_block = self._render_expectation_suite_notes(expectations)
        if asset_notes_content_block is not None:
            overview_content_blocks.append(asset_notes_content_block)

        sections = [
            RenderedSectionContent(**{
                "section_name": "Overview",
                "content_blocks": overview_content_blocks,
            })
        ]

        sections += [
            self._column_section_renderer.render(expectations=columns[column]) for column in ordered_columns if column != "_nocolumn"
        ]
        return RenderedDocumentContent(**{
            "renderer_type": "ExpectationSuitePageRenderer",
            "page_title": expectation_suite_name,
            "expectation_suite_name": expectation_suite_name,
            "utm_medium": "expectation-suite-page",
            "sections": sections
        })

    def _render_table_level_expectations(self, columns):
        table_level_expectations = columns.get("_nocolumn")
        if not table_level_expectations:
            return None
        else:
            expectation_bullet_list = self._column_section_renderer.render(
                expectations=table_level_expectations).content_blocks[1]
            expectation_bullet_list.header = RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Table-Level Expectations',
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            })
            return expectation_bullet_list

    @classmethod
    def _render_expectation_suite_header(cls):
        return RenderedHeaderContent(**{
            "content_block_type": "header",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Overview',
                    "tag": "h5",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "styling": {
                "classes": ["col-12"],
                "header": {
                    "classes": ["alert", "alert-secondary"]
                }
            }
        })

    @classmethod
    def _render_expectation_suite_info(cls, expectations):
        expectation_suite_name = expectations.expectation_suite_name
        ge_version = expectations.meta["great_expectations.__version__"]

        return RenderedTableContent(**{
            "content_block_type": "table",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Info',
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "table": [
                ["Expectation Suite Name", expectation_suite_name],
                ["Great Expectations Version", ge_version]
            ],
            "styling": {
                "classes": ["col-12", "table-responsive", "mt-1"],
                "body": {
                    "classes": ["table", "table-sm"]
                }
            },
        })

    # TODO: Update tests
    @classmethod
    def _render_expectation_suite_notes(cls, expectations):

        content = []

        total_expectations = len(expectations.expectations)
        columns = []
        for exp in expectations.expectations:
            if "column" in exp.kwargs:
                columns.append(exp.kwargs["column"])
        total_columns = len(set(columns))

        content = content + [
            # TODO: Leaving these two paragraphs as placeholders for later development.
            # "This Expectation suite was first generated by {BasicDatasetProfiler} on {date}, using version {xxx} of Great Expectations.",
            # "{name}, {name}, and {name} have also contributed additions and revisions.",
            "This Expectation suite currently contains %d total Expectations across %d columns." % (
                total_expectations,
                total_columns,
            ),
        ]

        if "notes" in expectations.meta:
            notes = expectations.meta["notes"]
            note_content = None

            if isinstance(notes, string_types):
                note_content = [notes]

            elif isinstance(notes, list):
                note_content = notes

            elif isinstance(notes, dict):
                if "format" in notes:
                    if notes["format"] == "string":
                        if isinstance(notes["content"], string_types):
                            note_content = [notes["content"]]
                        elif isinstance(notes["content"], list):
                            note_content = notes["content"]
                        else:
                            logger.warning("Unrecognized Expectation suite notes format. Skipping rendering.")

                    elif notes["format"] == "markdown":
                        if isinstance(notes["content"], string_types):
                            note_content = [
                                RenderedMarkdownContent(**{
                                    "content_block_type": "markdown",
                                    "markdown": notes["content"],
                                    "styling": {
                                        "parent": {
                                        }
                                    }
                                })
                            ]
                        elif isinstance(notes["content"], list):
                            note_content = [
                                RenderedMarkdownContent(**{
                                    "content_block_type": "markdown",
                                    "markdown": note,
                                    "styling": {
                                        "parent": {
                                        }
                                    }
                                }) for note in notes["content"]
                            ]
                        else:
                            logger.warning("Unrecognized Expectation suite notes format. Skipping rendering.")
                else:
                    logger.warning("Unrecognized Expectation suite notes format. Skipping rendering.")

            if note_content is not None:
                content = content + note_content

        return TextContent(**{
            "content_block_type": "text",
            "header": RenderedStringTemplateContent(**{
                "content_block_type": "string_template",
                "string_template": {
                    "template": 'Notes',
                    "tag": "h6",
                    "styling": {
                        "classes": ["m-0"]
                    }
                }
            }),
            "text": content,
            "styling": {
                "classes": ["col-12", "table-responsive", "mt-1"],
                "body": {
                    "classes": ["table", "table-sm"]
                }
            },
        })


class ProfilingResultsPageRenderer(Renderer):

    def __init__(self, overview_section_renderer=None, column_section_renderer=None):
        if overview_section_renderer is None:
            overview_section_renderer = {
                "class_name": "ProfilingResultsOverviewSectionRenderer"
            }
        if column_section_renderer is None:
            column_section_renderer = {
                "class_name": "ProfilingResultsColumnSectionRenderer"
            }
        module_name = 'great_expectations.render.renderer.other_section_renderer'
        self._overview_section_renderer = instantiate_class_from_config(
            config=overview_section_renderer,
            runtime_environment={},
            config_defaults={
                "module_name": overview_section_renderer.get("module_name", module_name)
            }
        )
        if not self._overview_section_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=overview_section_renderer['class_name']
            )
        module_name = 'great_expectations.render.renderer.column_section_renderer'
        self._column_section_renderer = instantiate_class_from_config(
            config=column_section_renderer,
            runtime_environment={},
            config_defaults={
                "module_name": column_section_renderer.get("module_name", module_name)
            }
        )
        if not self._column_section_renderer:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=column_section_renderer['class_name']
            )

    def render(self, validation_results):
        run_id = validation_results.meta['run_id']
        expectation_suite_name = validation_results.meta['expectation_suite_name']
        batch_kwargs = validation_results.meta.get("batch_kwargs")

        # add datasource key to batch_kwargs if missing
        if 'datasource' not in validation_results.meta.get("batch_kwargs", {}):
            # check if expectation_suite_name follows datasource.batch_kwargs_generator.data_asset_name.suite_name pattern
            if len(expectation_suite_name.split('.')) == 4:
                batch_kwargs['datasource'] = expectation_suite_name.split('.')[0]

        # Group EVRs by column
        # TODO: When we implement a ValidationResultSuite class, this method will move there.
        columns = self._group_evrs_by_column(validation_results)

        ordered_columns = Renderer._get_column_list_from_evrs(validation_results)
        column_types = self._overview_section_renderer._get_column_types(validation_results)

        return RenderedDocumentContent(**{
            "renderer_type": "ProfilingResultsPageRenderer",
            "page_title": run_id + "-" + expectation_suite_name + "-ProfilingResults",
            "expectation_suite_name": expectation_suite_name,
            "utm_medium": "profiling-results-page",
            "batch_kwargs": batch_kwargs,
            "sections":
                [
                    self._overview_section_renderer.render(
                        validation_results,
                        section_name="Overview"
                    )
                ] +
                [
                    self._column_section_renderer.render(
                        columns[column],
                        section_name=column,
                        column_type=column_types.get(column),
                    ) for column in ordered_columns
                ]
        })



import json
from pprint import pformat

import os
from airflow import AirflowException
from airflow.hooks.S3_hook import S3Hook
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.dbapi_hook import DbApiHook
from airflow.models import BaseOperator
import great_expectations as ge
from examples.integrations.airflow.hooks.s3_csv_hook import ExpectationS3CsvHook
from examples.integrations.airflow.hooks.db_hook import ExpectationMySQLHook

####
#
# NOTE: this code has not been updated for the new GE 0.7 naming conventions
#
####

class ExpectationOperator(BaseOperator):

    results_dest_name = None
    template_fields = ['results_dest_name']

    def __init__(self,
                 dataset,
                 expectations_json,
                 fail_on_error=True,
                 source_conn_id=None,
                 dest_conn_id=None,
                 source_bucket_name=None,
                 results_dest_name=None,
                 results_bucket_name=None,
                 dataset_params=None,
                 *args, **kwargs):
        """
        Validate provided dataset using great_expectations.
        :param dataset: Name of the dataset being loaded
        :type str
        :param expectations_json: file pointing to expectation suite or json string
        :type str
        :param fail_on_error: True if airflow job should fail when expectations fail
        :type bool
        :param conn_type: Type of conn_id that's being provided. (e.g s3connection or MySQL). Necessary for deciding
        which hook to use
        :type str
        :param source_conn_id: Connection that should be used to load dataset
        :type str
        :param source_bucket_name: Name of bucket where dataset can be found (only when using s3_csv connection)
        :type str
        :param dataset_params: Any arguments that need to be passed to great_expectations validate method
        :type dict
        :param results_dest_name: Name of file where results will be stored
        :type str
        :param results_bucket_name: Name of s3 bucket where results will be stored
        :type str
        """
        super().__init__(*args, **kwargs)

        # Get source hook
        if source_conn_id is not None:
            self._setup_source_conn(source_conn_id, source_bucket_name)

        # Get destination hook and make sure all required parameters are set
        if dest_conn_id is not None:
            self._setup_dest_conn(dest_conn_id, results_bucket_name, results_dest_name)

        self.expectations_json = expectations_json
        self.fail_on_error = fail_on_error

        if dataset_params is None:
            dataset_params = {}

        self.dataset_params = dataset_params
        self.dataset_name = dataset

    def _setup_source_conn(self, source_conn_id, source_bucket_name=None):
        """
        Retrieve connection based on source_conn_id. In case of s3 it also configures the bucket.
        Validates that connection id belongs to supported connection type.
        :param source_conn_id:
        :param source_bucket_name:
        """
        self.source_conn = BaseHook.get_hook(source_conn_id)
        self.source_conn_id = source_conn_id

        # Workaround for getting hook in case of s3 connection
        # This is needed because get_hook silently returns None for s3 connections
        # See https://issues.apache.org/jira/browse/AIRFLOW-2316 for more info
        connection = BaseHook._get_connection_from_env(source_conn_id)
        self.log.info(connection.extra_dejson)
        if connection.conn_type == 's3':
            self.log.info("Setting up s3 connection {0}".format(source_conn_id))
            self.source_conn = S3Hook(aws_conn_id=source_conn_id)
            # End Workaround
            if source_bucket_name is None:
                raise AttributeError("Missing source bucket for s3 connection")
            self.source_bucket_name = source_bucket_name

        if not isinstance(self.source_conn, DbApiHook) and not isinstance(self.source_conn, S3Hook):
            raise AttributeError(
                "Only s3_csv, local and sql connection types are allowed, not {0}".format(type(self.source_conn)))

    def _setup_dest_conn(self, dest_conn_id, results_bucket_name, results_dest_name):
        """
        Setup results connection. Retrieves s3 connection and makes sure we've got location details (bucket, filename)
        :param dest_conn_id:
        :param results_bucket_name:
        :param results_dest_name:
        """
        conn = BaseHook._get_connection_from_env(dest_conn_id)
        if conn.conn_type != 's3':
            raise AttributeError(
                "Only s3 is allowed as a results destination, not {0}".format(conn.conn_type))

        self.dest_conn = S3Hook(aws_conn_id=dest_conn_id)
        self.dest_conn_id = dest_conn_id

        if results_bucket_name is None or results_dest_name is None:
            raise AttributeError("Specify bucket name and key name to store results")

        self.results_bucket_name = results_bucket_name
        self.results_dest_name = results_dest_name

    def _get_dataframe(self):
        """
        Load dataframe based on specified connection
        :return:
        """
        if self.source_conn is None:  # Use local file
            return ge.read_csv(self.dataset_name, **self.dataset_params)

        if isinstance(self.source_conn, S3Hook):
            hook = ExpectationS3CsvHook(aws_conn_id=self.source_conn_id)

            return hook.get_ge_df(self.dataset_name, self.source_bucket_name, **self.dataset_params)

        if isinstance(self.source_conn, DbApiHook):
            hook = ExpectationMySQLHook(mysql_conn_id=self.source_conn_id)

            return hook.get_ge_df(self.dataset_name, **self.dataset_params)

    def _load_json(self):
        """
        Load expectation suite based on operator parameters. If provided expectations_json is a file the suite will
        be loaded from this file. Otherwise we'll try to load the expectation suite as a string.
        :return:
        """
        if os.path.isfile(self.expectations_json):
            self.log.info("Loading expectation suite from file {file}".format(file=self.expectations_json))
            return json.load(open(self.expectations_json))
        else:
            self.log.info("Loading expectation suite from string")
            return json.loads(self.expectations_json)

    def _store_results(self, results):
        hook = ExpectationS3CsvHook(aws_conn_id=self.dest_conn_id)
        return hook.dump_results(results, self.results_bucket_name, self.results_dest_name)

    def execute(self, context):
        df = self._get_dataframe()
        suite = self._load_json()
        self.log.info("Start dataset validation for set {set}".format(set=self.dataset_name))
        results = df.validate(expectation_suite=suite)

        self.log.info(pformat(results))

        if self.dest_conn is not None:
            self._store_results(results)

        for result in results['results']:
            if result.success is False:
                if self.fail_on_error is True:
                    raise AirflowException("Validation failed for dataset {name}".format(name=self.dataset_name))

        return results



import pandas as pd
import numpy as np
import scipy.stats as stats
import great_expectations as ge
import json
import sys
import copy

"""
Use this file to generate random datasets for testing distributional expectations.

Tests expect two datasets: "distributional_expectations_data_base.csv" and "distributional_expectations_data_test.csv" 
They also expect a set of partitions: "test_partitions.json"

The partitions should be built from distributional_expectations_data_base.csv. The tests will use distributional_expectations_data_test.csv

"""


def generate_new_data(seed):
    np.random.seed(seed=seed)

    norm_0_1 = stats.norm.rvs(0, 1, 1000)
    norm_1_1 = stats.norm.rvs(1, 1, 1000)
    norm_10_1 = stats.norm.rvs(10, 1, 1000)
    bimodal = np.concatenate((norm_0_1[:500], norm_10_1[500:]))
    categorical_fixed = (['A'] * 540) + (['B'] * 320) + (['C'] * 140)

    return pd.DataFrame({
        'norm_0_1': norm_0_1,
        'norm_1_1': norm_1_1,
        'norm_10_1': norm_10_1,
        'bimodal': bimodal,
        'categorical_fixed': categorical_fixed
    })


def generate_new_partitions(df):
    test_partitions = {}
    for column in ['norm_0_1', 'norm_1_1', 'bimodal']:
        partition_object = ge.dataset.util.kde_partition_data(df[column])
        # Print how close sum of weights is to one for a quick visual consistency check when data are generated
        #print(column + '_kde: '+ str(abs(1-np.sum(partition_object['weights']))))
        test_partitions[column + '_kde'] = partition_object

        for bin_type in ['uniform', 'ntile', 'auto']:
            partition_object = ge.dataset.util.continuous_partition_data(
                df[column], bin_type)
            # Print how close sum of weights is to one for a quick visual consistency check when data are generated
            #print(column + '_' + bin_type + ': ' + str(abs(1 - np.sum(partition_object['weights']))))
            test_partitions[column + '_' + bin_type] = partition_object

        # Create infinite endpoint partitions:
        inf_partition = copy.deepcopy(test_partitions[column + "_auto"])
        inf_partition["weights"] = inf_partition["weights"] * (1 - 0.01)
        inf_partition["tail_weights"] = [0.005, 0.005]
        test_partitions[column + '_auto_inf'] = inf_partition

    partition_object = ge.dataset.util.categorical_partition_data(
        df['categorical_fixed'])
    test_partitions['categorical_fixed'] = partition_object
    alt_partition = ge.dataset.util.categorical_partition_data(
        df['categorical_fixed'])
    # overwrite weights with uniform weights to give a testing dataset
    alt_partition['weights'] = [
        1./len(alt_partition['values'])] * len(alt_partition['values'])
    test_partitions['categorical_fixed_alternate'] = alt_partition

    return test_partitions


if __name__ == "__main__":
    df = generate_new_data(seed=42)
    d = df.to_dict(orient="list")
    json.dump(d, open('../test_sets/distributional_expectations_data_base.json', 'w'))
    test_partitions = generate_new_partitions(df)

    test_partitions = ge.data_asset.util.recursively_convert_to_json_serializable(test_partitions)
    with open('../test_sets/test_partitions_definition_fixture.json', 'w') as file:
        file.write(json.dumps(test_partitions))

    df = generate_new_data(seed=20190501)
    d = df.to_dict(orient="list")
    json.dump(d, open('../test_sets/distributional_expectations_data_test.json', 'w'))
    print("Done generating new base data, partitions, and test data.")



import os

import nbformat
import pytest
from nbconvert.preprocessors import ExecutePreprocessor

from great_expectations import DataContext
from great_expectations.exceptions import DataContextError
from great_expectations.render.renderer.suite_scaffold_notebook_renderer import (
    SuiteScaffoldNotebookRenderer,
)


def test_render_snapshot_test(titanic_data_context):
    batch_kwargs = titanic_data_context.build_batch_kwargs(
        "mydatasource", "mygenerator", "Titanic"
    )
    csv_path = batch_kwargs["path"]
    renderer = SuiteScaffoldNotebookRenderer(
        titanic_data_context, "my_suite", batch_kwargs
    )
    obs = renderer.render(None, None)
    assert isinstance(obs, nbformat.NotebookNode)
    ## NOTE!!! - When updating this snapshot be sure to include the dynamic
    # csv_path in the second cell due to pytest fixtures
    expected = {
        "nbformat": 4,
        "nbformat_minor": 4,
        "metadata": {},
        "cells": [
            {
                "cell_type": "markdown",
                "source": "# Scaffold a new Expectation Suite (BETA)\nUse this notebook to scaffold a new expectations suite. This process helps you\navoid writing lots of boilerplate when authoring suites.\n\n**Expectation Suite Name**: `my_suite`\n\nWe'd love it if you **reach out to us on** the [**Great Expectations Slack Channel**](https://greatexpectations.io/slack)",
                "metadata": {},
            },
            {
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'from datetime import datetime\nimport great_expectations as ge\nimport great_expectations.jupyter_ux\nfrom great_expectations.profile import BasicSuiteBuilderProfiler\nfrom great_expectations.data_context.types.resource_identifiers import (\n    ValidationResultIdentifier,\n)\n\ncontext = ge.data_context.DataContext()\n\nexpectation_suite_name = "my_suite"\nsuite = context.create_expectation_suite(\n    expectation_suite_name, overwrite_existing=True\n)\n\nbatch_kwargs = {\n    "path": "'
                + csv_path
                + '",\n    "datasource": "mydatasource",\n}\nbatch = context.get_batch(batch_kwargs, suite)\nbatch.head()',
                "outputs": [],
            },
            {
                "cell_type": "markdown",
                "source": "## Select the columns you want to scaffold expectations on\n\nSimply uncomment columns that are important. You can select multiple lines and\nuse a jupyter keyboard shortcut to toggle each line: **Linux/Windows**:\n`Ctrl-/`, **macOS**: `Cmd-/`",
                "metadata": {},
            },
            {
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": "included_columns = [\n    # 'Unnamed: 0',\n    # 'Name',\n    # 'PClass',\n    # 'Age',\n    # 'Sex',\n    # 'Survived',\n    # 'SexCode'\n]",
                "outputs": [],
            },
            {
                "cell_type": "markdown",
                "source": "## Run the scaffolder\n\nThis is highly configurable depending on your goals. You can include or exclude\ncolumns, and include or exclude expectation types (when applicable). [The \nExpectation Glossary](http://docs.greatexpectations.io/en/latest/expectation_glossary.html) \ncontains a list of possible expectations.\n\nNote that the profiler is not very smart, so it does it's best to decide on\napplicability.\n\n**To get to a production grade suite, you should [edit this \nsuite](http://docs.greatexpectations.io/en/latest/command_line.html#great-expectations-suite-edit) \nafter this scaffold gets you close to what you want.**",
                "metadata": {},
            },
            {
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": '# Wipe the suite clean to prevent unwanted expectations on the batch\nsuite = context.create_expectation_suite(expectation_suite_name, overwrite_existing=True)\nbatch = context.get_batch(batch_kwargs, suite)\n\nscaffold_config = {\n    "included_columns": included_columns,\n    # "excluded_columns": [],\n    # "included_expectations: [],\n    # "excluded_expectations: [],\n}\nsuite, evr = BasicSuiteBuilderProfiler().profile(batch, profiler_configuration=scaffold_config)',
                "outputs": [],
            },
            {
                "cell_type": "markdown",
                "source": "## Save & review the scaffolded Expectation Suite\n\nLet's save the scaffolded expectation suite as a JSON file in the \n`great_expectations/expectations` directory of your project and rebuild the Data\n Docs site to make reviewing the scaffolded suite easy.",
                "metadata": {},
            },
            {
                "cell_type": "code",
                "metadata": {},
                "execution_count": None,
                "source": 'context.save_expectation_suite(suite, expectation_suite_name)\n\n# Let\'s make a simple sortable timestamp. Note this could come from your pipeline runner.\nrun_id = datetime.utcnow().strftime("%Y%m%dT%H%M%S.%fZ")\n\nresults = context.run_validation_operator("action_list_operator", assets_to_validate=[batch], run_id=run_id)\nexpectation_suite_identifier = list(results["details"].keys())[0]\nvalidation_result_identifier = ValidationResultIdentifier(\n    expectation_suite_identifier=expectation_suite_identifier,\n    batch_identifier=batch.batch_kwargs.to_id(),\n    run_id=run_id\n)\ncontext.build_data_docs()\ncontext.open_data_docs(validation_result_identifier)',
                "outputs": [],
            },
            {
                "cell_type": "markdown",
                "source": "## Next steps\nAfter you are happy with this scaffolded Expectation Suite in Data Docs you \nshould edit this suite to make finer grained adjustments to the expectations. \nThis is be done by running `great_expectations suite edit my_suite`.",
                "metadata": {},
            },
        ],
    }
    del expected["nbformat_minor"]
    del obs["nbformat_minor"]

    for obs_cell, expected_cell in zip(obs["cells"], expected["cells"]):
        assert obs_cell == expected_cell
    assert obs == expected


def test_notebook_execution_with_pandas_backend(titanic_data_context):
    """
    This tests that the notebook is written to disk and executes without error.

    To set this test up we:
    - create a scaffold notebook
    - verify that no validations have happened

    We then:
    - execute that notebook (Note this will raise various errors like
    CellExecutionError if any cell in the notebook fails
    - create a new context from disk
    - verify that a validation has been run with our expectation suite
    """
    context = titanic_data_context
    root_dir = context.root_directory
    uncommitted_dir = os.path.join(root_dir, "uncommitted")
    suite_name = "my_suite"

    csv_path = os.path.join(root_dir, "..", "data", "Titanic.csv")
    batch_kwargs = {"datasource": "mydatasource", "path": csv_path}

    # Sanity check test setup
    assert context.list_expectation_suite_names() == []
    assert context.list_datasources() == [
        {
            "module_name": "great_expectations.datasource",
            "class_name": "PandasDatasource",
            "data_asset_type": {
                "module_name": "great_expectations.dataset",
                "class_name": "PandasDataset",
            },
            "batch_kwargs_generators": {
                "mygenerator": {
                    "class_name": "SubdirReaderBatchKwargsGenerator",
                    "base_directory": "../data",
                }
            },
            "name": "mydatasource",
        }
    ]
    assert context.get_validation_result(suite_name) == {}
    notebook_path = os.path.join(uncommitted_dir, f"{suite_name}.ipynb")
    assert not os.path.isfile(notebook_path)

    # Create notebook
    renderer = SuiteScaffoldNotebookRenderer(
        titanic_data_context, suite_name, batch_kwargs
    )
    renderer.render_to_disk(notebook_path)
    assert os.path.isfile(notebook_path)

    with open(notebook_path, "r") as f:
        nb = nbformat.read(f, as_version=4)

    # Run notebook
    ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
    ep.preprocess(nb, {"metadata": {"path": uncommitted_dir}})

    # Useful to inspect executed notebook
    output_notebook = os.path.join(uncommitted_dir, "output.ipynb")
    with open(output_notebook, "w") as f:
        nbformat.write(nb, f)

    # Assertions about output
    context = DataContext(root_dir)
    obs_validation_result = context.get_validation_result(suite_name)
    assert obs_validation_result.statistics == {
        "evaluated_expectations": 3,
        "successful_expectations": 3,
        "unsuccessful_expectations": 0,
        "success_percent": 100,
    }
    suite = context.get_expectation_suite(suite_name)
    assert suite.expectations


def test_notebook_rendering_throws_error_on_existing_suite(
    titanic_data_context,
):
    """
    To set this test up we:
    - create a suite

    We then:
    - create a scaffold notebook for the same suite and expect an error during rendering
    """
    context = titanic_data_context
    root_dir = context.root_directory
    uncommitted_dir = os.path.join(root_dir, "uncommitted")
    suite_name = "my_suite"

    csv_path = os.path.join(root_dir, "..", "data", "Titanic.csv")
    batch_kwargs = {"datasource": "mydatasource", "path": csv_path}
    suite = context.create_expectation_suite(suite_name)
    context.save_expectation_suite(suite)

    # Sanity check test setup
    assert context.list_expectation_suite_names() == [suite_name]
    notebook_path = os.path.join(uncommitted_dir, f"{suite_name}.ipynb")
    assert not os.path.isfile(notebook_path)

    # Create notebook
    with pytest.raises(DataContextError):
        renderer = SuiteScaffoldNotebookRenderer(
            titanic_data_context, suite_name, batch_kwargs
        )
        renderer.render_to_disk(notebook_path)



from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer
from great_expectations.core.usage_statistics.anonymizers.store_backend_anonymizer import StoreBackendAnonymizer
from great_expectations.core.usage_statistics.anonymizers.site_builder_anonymizer import SiteBuilderAnonymizer


class DataDocsSiteAnonymizer(Anonymizer):
    def __init__(self, salt=None):
        super(DataDocsSiteAnonymizer, self).__init__(salt=salt)
        self._site_builder_anonymizer = SiteBuilderAnonymizer(salt=salt)
        self._store_backend_anonymizer = StoreBackendAnonymizer(salt=salt)

    def anonymize_data_docs_site_info(self, site_name, site_config):
        site_config_module_name = site_config.get("module_name")
        if site_config_module_name is None:
            site_config["module_name"] = "great_expectations.render.renderer.site_builder"

        anonymized_info_dict = self._site_builder_anonymizer.anonymize_site_builder_info(
            site_builder_config=site_config,
        )
        anonymized_info_dict["anonymized_name"] = self.anonymize(site_name)

        store_backend_config = site_config.get("store_backend")
        anonymized_info_dict["anonymized_store_backend"] = self._store_backend_anonymizer.anonymize_store_backend_info(
            store_backend_object_config=store_backend_config
        )
        site_index_builder_config = site_config.get("site_index_builder")
        anonymized_site_index_builder = self._site_builder_anonymizer.anonymize_site_builder_info(
            site_builder_config=site_index_builder_config
        )
        if "show_cta_footer" in site_index_builder_config:
            anonymized_site_index_builder["show_cta_footer"] = site_index_builder_config.get("show_cta_footer")
        anonymized_info_dict["anonymized_site_index_builder"] = anonymized_site_index_builder

        return anonymized_info_dict